(function () {
    console.log("PinkJSBridge initialization begin");
    if (window.PinkJSBridge) {
        return;
    }

    var vsprintf = function (fmt, argv) {
        argv.unshift(fmt);
        return sprintf.apply(null, argv);
    };

    function printObject(o) {
        var out = '';
        for (var p in o) {
            out += p + ': ' + o[p] + '\n';
        }
        return out;
    }

    var UTF8 = {
        encode: function (string) {
            string = string.replace(/\r\n/g, "\n");
            var utftext = "";

            for (var n = 0; n < string.length; n++) {
                var c = string.charCodeAt(n);
                if (c < 128) {
                    utftext += String.fromCharCode(c);
                } else if ((c > 127) && (c < 2048)) {
                    utftext += String.fromCharCode((c >> 6) | 192);
                    utftext += String.fromCharCode((c & 63) | 128);
                } else {
                    utftext += String.fromCharCode((c >> 12) | 224);
                    utftext += String.fromCharCode(((c >> 6) & 63) | 128);
                    utftext += String.fromCharCode((c & 63) | 128);
                }
            }
            return utftext;
        },

        decode: function (utftext) {
            var string = "";
            var i = 0;
            var c = c1 = c2 = 0;
            while (i < utftext.length) {
                c = utftext.charCodeAt(i);
                if (c < 128) {
                    string += String.fromCharCode(c);
                    i++;
                } else if ((c > 191) && (c < 224)) {
                    c2 = utftext.charCodeAt(i + 1);
                    string += String.fromCharCode(((c & 31) << 6) | (c2 & 63));
                    i += 2;
                } else {
                    c2 = utftext.charCodeAt(i + 1);
                    c3 = utftext.charCodeAt(i + 2);
                    string += String.fromCharCode(((c & 15) << 12) | ((c2 & 63) << 6) | (c3 & 63));
                    i += 3;
                }
            }
            return string;
        }
    };

    var HEX = {
        encode: function (str) {
            var r = "";
            var e = str.length;
            var c = 0;
            var h;
            while (c < e) {
                h = str.charCodeAt(c++).toString(16);
                while (h.length < 2) h = "0" + h;
                r += h;
            }
            return r.toUpperCase();
        },
        decode: function (str) {
            var r = "";
            var e = str.length;
            var s;
            while (e >= 0) {
                s = e - 2;
                r = String.fromCharCode("0x" + str.substring(s, e)) + r;
                e = s;
            }
            return r;
        }
    };

    var JSON;
    if (!JSON) {
        JSON = {};
    }
    (function () {
        'use strict';
        function f(n) {
            return n < 10 ? '0' + n : n;
        }

        if (typeof Date.prototype.toJSON !== 'function') {
            Date.prototype.toJSON = function (key) {
                return isFinite(this.valueOf()) ? this.getUTCFullYear() + '-' + f(this.getUTCMonth() + 1) + '-' + f(this.getUTCDate()) + 'T' + f(this.getUTCHours()) + ':' + f(this.getUTCMinutes()) + ':' + f(this.getUTCSeconds()) + 'Z' : null;
            };
            String.prototype.toJSON = Number.prototype.toJSON = Boolean.prototype.toJSON = function (key) {
                return this.valueOf();
            };
        }
        var cx = /[\u0000\u00ad\u0600-\u0604\u070f\u17b4\u17b5\u200c-\u200f\u2028-\u202f\u2060-\u206f\ufeff\ufff0-\uffff]/g,
            escapable = /[\\\"\x00-\x1f\x7f-\x9f\u00ad\u0600-\u0604\u070f\u17b4\u17b5\u200c-\u200f\u2028-\u202f\u2060-\u206f\ufeff\ufff0-\uffff]/g,
            gap, indent, meta = {
                '\b': '\\b',
                '\t': '\\t',
                '\n': '\\n',
                '\f': '\\f',
                '\r': '\\r',
                '"': '\\"',
                '\\': '\\\\'
            },
            rep;

        function quote(string) {
            escapable.lastIndex = 0;
            return escapable.test(string) ? '"' + string.replace(escapable, function (a) {
                var c = meta[a];
                return typeof c === 'string' ? c : '\\u' + ('0000' + a.charCodeAt(0).toString(16)).slice(-4);
            }) + '"' : '"' + string + '"';
        }

        function str(key, holder) {
            var i,
                k,
                v,
                length, mind = gap, partial, value = holder[key];
            if (typeof rep === 'function') {
                value = rep.call(holder, key, value);
            }

            switch (typeof value) {
                case 'string':
                    return quote(value);
                case 'number':
                    return isFinite(value) ? String(value) : 'null';
                case 'boolean':
                case 'null':
                    return String(value);
                case 'object':
                    if (!value) {
                        return 'null';
                    }
                    gap += indent;
                    partial = [];

                    if (Object.prototype.toString.apply(value) === '[object Array]') {
                        length = value.length;
                        for (i = 0; i < length; i += 1) {
                            partial[i] = str(i, value) || 'null';
                        }
                        v = partial.length === 0 ? '[]' : gap ? '[\n' + gap + partial.join(',\n' + gap) + '\n' + mind + ']' : '[' + partial.join(',') + ']';
                        gap = mind;
                        return v;
                    }

                    if (rep && typeof rep === 'object') {
                        length = rep.length;
                        for (i = 0; i < length; i += 1) {
                            if (typeof rep[i] === 'string') {
                                k = rep[i];
                                v = str(k, value);
                                if (v) {
                                    partial.push(quote(k) + (gap ? ': ' : ':') + v);
                                }
                            }
                        }
                    } else {
                        for (k in value) {
                            if (Object.prototype.hasOwnProperty.call(value, k)) {
                                v = str(k, value);
                                if (v) {
                                    partial.push(quote(k) + (gap ? ': ' : ':') + v);
                                }
                            }
                        }
                    }
                    v = partial.length === 0 ? '{}' : gap ? '{\n' + gap + partial.join(',\n' + gap) + '\n' + mind + '}' : '{' + partial.join(',') + '}';
                    gap = mind;
                    return v;
            }
        }

        if (typeof JSON.stringify !== 'function') {
            JSON.stringify = function (value, replacer, space) {
                var i;
                gap = '';
                indent = '';
                if (typeof space === 'number') {
                    for (i = 0; i < space; i += 1) {
                        indent += ' ';
                    }
                } else if (typeof space === 'string') {
                    indent = space;
                }
                rep = replacer;
                if (replacer && typeof replacer !== 'function' && (typeof replacer !== 'object' || typeof replacer.length !== 'number')) {
                    throw new Error('JSON.stringify');
                }
                return str('', {
                    '': value
                });
            };
        }

        if (typeof JSON.parse !== 'function') {
            JSON.parse = function (text, reviver) {
                var j;

                function walk(holder, key) {
                    var k, v, value = holder[key];
                    if (value && typeof value === 'object') {
                        for (k in value) {
                            if (Object.prototype.hasOwnProperty.call(value, k)) {
                                v = walk(value, k);
                                if (v !== undefined) {
                                    value[k] = v;
                                } else {
                                    delete value[k];
                                }
                            }
                        }
                    }
                    return reviver.call(holder, key, value);
                }

                text = String(text);
                cx.lastIndex = 0;
                if (cx.test(text)) {
                    text = text.replace(cx, function (a) {
                        return '\\u' + ('0000' + a.charCodeAt(0).toString(16)).slice(-4);
                    });
                }
                if (/^[\],:{}\s]*$/.test(text.replace(/\\(?:["\\\/bfnrt]|u[0-9a-fA-F]{4})/g, '@').replace(/"[^"\\\n\r]*"|true|false|null|-?\d+(?:\.\d*)?(?:[eE][+\-]?\d+)?/g, ']').replace(/(?:^|:|,)(?:\s*\[)+/g, ''))) {
                    j = eval('(' + text + ')');
                    return typeof reviver === 'function' ? walk({
                        '': j
                    }, '') : j;
                }
                throw new SyntaxError('JSON.parse');
            };
        }
    }());

    var base64encodechars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

    function base64encode(str) {
        if (str === undefined) {
            return str;
        }

        var out, i, len;
        var c1, c2, c3;
        len = str.length;
        i = 0;
        out = "";
        while (i < len) {
            c1 = str.charCodeAt(i++) & 0xff;
            if (i == len) {
                out += base64encodechars.charAt(c1 >> 2);
                out += base64encodechars.charAt((c1 & 0x3) << 4);
                out += "==";
                break;
            }
            c2 = str.charCodeAt(i++);
            if (i == len) {
                out += base64encodechars.charAt(c1 >> 2);
                out += base64encodechars.charAt(((c1 & 0x3) << 4) | ((c2 & 0xf0) >> 4));
                out += base64encodechars.charAt((c2 & 0xf) << 2);
                out += "=";
                break;
            }
            c3 = str.charCodeAt(i++);
            out += base64encodechars.charAt(c1 >> 2);
            out += base64encodechars.charAt(((c1 & 0x3) << 4) | ((c2 & 0xf0) >> 4));
            out += base64encodechars.charAt(((c2 & 0xf) << 2) | ((c3 & 0xc0) >> 6));
            out += base64encodechars.charAt(c3 & 0x3f);
        }
        return out;
    }

    var messagingIframe, setResultIframe, receiveMessageQueue = [], sendMessageQueue = [], messageHandlers = {};
    var responseCallbacks = {};
    var eventsCallbacks = {};
    var uniqueId = 1;

    var CUSTOM_PROTOCOL_SCHEME = 'pink';
    var QUEUE_HAS_MESSAGE = 'dispatch_message/';

    function _createQueueReadyIframe(doc) {
        console.log("[_createQueueReadyIframe]");
        setResultIframe = doc.createElement("iframe");
        setResultIframe.id = "__PinkJSBridgeIframe_SetResult";
        setResultIframe.style.display = "none";
        doc.documentElement.appendChild(setResultIframe);

        messagingIframe = doc.createElement("iframe");
        messagingIframe.id = "__PinkJSBridgeIframe";
        messagingIframe.style.display = "none";
        doc.documentElement.appendChild(messagingIframe);
    }

    function _dispatchMessageFromApp(messageJSON) {
        console.log("_dispatchMessageFromApp: "+ messageJSON);
        setTimeout(function _timeoutDispatchMessageFromApp() {
            var message = messageJSON;
            if(typeof messageJSON === "string"){
                message = JSON.parse(messageJSON)
            }
            var messageHandler;
            console.log("message.responseId: "+ message.responseId);
            if (message.responseId) {
                var responseCallback = responseCallbacks[message.responseId];
                console.log("responseCallback: "+ responseCallback);
                if (!responseCallback) { return; }
                responseCallback(message.responseData);
                console.log("message.responseData: "+ message.responseData);
                if(message.responseId.search("once") != -1) {
                    delete responseCallbacks[message.responseId];
                }
            } else {
                var responseCallback;
                if (message.callbackId) {
                    var callbackResponseId = message.callbackId;
                    responseCallback = function(responseData) {
                        _doSend({ responseId:callbackResponseId, responseData:responseData });
                    };
                }

                var handler = PinkJSBridge._messageHandler;
                if (message.handlerName) {
                    handler = messageHandlers[message.handlerName];
                }

                try {
                    handler(message.data, responseCallback);
                } catch(exception) {
                    if (typeof console != 'undefined') {
                        console.log("PinkJSBridge: WARNING: javascript handler threw.", message, exception);
                    }
                }
            }
        });
    }

    function _doSend(message, responseCallback) {
        if (responseCallback) {
            var callbackId = 'cb_' + (uniqueId++) + '_' + new Date().getTime();
            if(message['once']){
                callbackId = callbackId + "_once";
            }else{
                var handlerName = message['handlerName'];
                eventsCallbacks[handlerName] = callbackId;
            }
            responseCallbacks[callbackId] = responseCallback;
            message['callbackId'] = callbackId;
        }
        sendMessageQueue.push(message);
        console.log("[_doSend]sendMessageQueue.push:" + message);
        var src = CUSTOM_PROTOCOL_SCHEME + '://' + QUEUE_HAS_MESSAGE;
        prompt({}, src);
    }

    function init(messageHandler) {
        console.log("[init]messageHandler=" + messageHandler);
        if (PinkJSBridge._messageHandler) {
            throw new Error('PinkJSBridge.init called twice');
        }
        PinkJSBridge._messageHandler = messageHandler;
        var receivedMessages = receiveMessageQueue;
        receiveMessageQueue = null;
        for (var i = 0; i < receivedMessages.length; i++) {
            _dispatchMessageFromApp(receivedMessages[i]);
        }
    }

    function send(data, responseCallback) {
        console.log("[send]data=" + data + ",callback=" + responseCallback);
        _doSend({
            data: data
        }, responseCallback);
    }

    function registerHandler(handlerName, handler) {
        messageHandlers[handlerName] = handler
    }

    function callHandler(handlerName, data, responseCallback) {
        console.log("[callHandler]handlerName=" + handlerName + ",data="
            + data + ",responseCallback=" + responseCallback);
        var once = true;
        if(handlerName == "onShow" || handlerName == "onHide"){
            once = false;
        }
        if(handlerName == "offShow"){
            var cbId = eventsCallbacks["onShow"];
            if(cbId && responseCallbacks[cbId]){
                delete responseCallbacks[cbId];
            }
        }
        if(handlerName == "offHide"){
            var cbId = eventsCallbacks["onHide"];
            if(cbId && responseCallbacks[cbId]){
                delete responseCallbacks[cbId];
            }
        }
        _doSend({ handlerName: handlerName, data: data, once: once }, responseCallback);
    }

    function _log(fmt) {
        console.log("[_log]fmt=" + fmt);
        var argv = [];
        for (var i = 0; i < arguments.length; i++) {
            argv.push(arguments[i]);
        }

        var fm = argv.shift();
        var msg;
        try {
            msg = vsprintf(fm, argv);
        } catch (e) {
            msg = fmt;
        }
        callHandler('log', {
            'msg': msg
        });
    }

    function _setResultValue(scene, result) {
        if (result === undefined) {
            result = 'dummy';
        }
        var src = 'pink://private/setresult/' + scene + '&' + base64encode(UTF8.encode(result));
        prompt({}, src);
    }

    function _fetchQueue() {
        console.log("[_fetchQueue]");
        var messageQueueString = JSON.stringify(sendMessageQueue);
        sendMessageQueue = [];
        _setResultValue('SCENE_FETCHQUEUE', messageQueueString);
    }

    function _handleMessageFromPink(message) {
        console.log("[_handleMessageFromPink]message=" + message);
        _dispatchMessageFromApp(message);
    }

    var _PinkJSBridge = {
        init: init,
        send: send,
        registerHandler: registerHandler,
        callHandler: callHandler,
        log: _log,
        _fetchQueue: _fetchQueue,
        _handleMessageFromPink: _handleMessageFromPink,
        _hasInit: false
    };
    window.PinkJSBridge = _PinkJSBridge;

    var doc = document;
    _createQueueReadyIframe(doc);
    var readyEvent = doc.createEvent('Events');
    readyEvent.initEvent('PinkJSBridgeReady');
    readyEvent.bridge = window.PinkJSBridge;
    doc.dispatchEvent(readyEvent);

    console.log("PinkJSBridge initialization end")
})();